import adjacency_graphs from "./adjacency_graphs.mjs";
const calc_average_degree = function(graph) {
  let average = 0;
  for (const key in graph) {
    const neighbors = graph[key];
    average += function() {
      const results = [];
      for (let o = 0, len = neighbors.length; o < len; o++) {
        const n = neighbors[o];
        if (n) {
          results.push(n);
        }
      }
      return results;
    }().length;
  }
  average /= function() {
    const results = [];
    for (const k in graph) {
      const v = graph[k];
      results.push(k);
    }
    return results;
  }().length;
  return average;
};
const BRUTEFORCE_CARDINALITY = 10;
const MIN_GUESSES_BEFORE_GROWING_SEQUENCE = 1e4;
const MIN_SUBMATCH_GUESSES_SINGLE_CHAR = 10;
const MIN_SUBMATCH_GUESSES_MULTI_CHAR = 50;
const scoring = {
  nCk: function(n, k) {
    if (k > n) {
      return 0;
    }
    if (k === 0) {
      return 1;
    }
    let r = 1;
    let d, o, ref;
    for (d = o = 1, ref = k; 1 <= ref ? o <= ref : o >= ref; d = 1 <= ref ? ++o : --o) {
      r *= n;
      r /= d;
      n -= 1;
    }
    return r;
  },
  log10: function(n) {
    return Math.log(n) / Math.log(10);
  },
  log2: function(n) {
    return Math.log(n) / Math.log(2);
  },
  factorial: function(n) {
    if (n < 2) {
      return 1;
    }
    let f = 1;
    let i, o, ref;
    for (i = o = 2, ref = n; 2 <= ref ? o <= ref : o >= ref; i = 2 <= ref ? ++o : --o) {
      f *= i;
    }
    return f;
  },
  most_guessable_match_sequence: function(password, matches, _exclude_additive = false) {
    const n = password.length;
    const matches_by_j = function() {
      const results = [];
      let _, ref2, o;
      for (_ = o = 0, ref2 = n; 0 <= ref2 ? o < ref2 : o > ref2; _ = 0 <= ref2 ? ++o : --o) {
        results.push([]);
      }
      return results;
    }();
    for (let o = 0, len = matches.length; o < len; o++) {
      const m = matches[o];
      matches_by_j[m.j].push(m);
    }
    for (let q = 0, len1 = matches_by_j.length; q < len1; q++) {
      const lst = matches_by_j[q];
      lst.sort(function(m1, m2) {
        return m1.i - m2.i;
      });
    }
    const optimal = {
      m: function() {
        const results = [];
        let _, u2, ref2;
        for (_ = u2 = 0, ref2 = n; 0 <= ref2 ? u2 < ref2 : u2 > ref2; _ = 0 <= ref2 ? ++u2 : --u2) {
          results.push({});
        }
        return results;
      }(),
      pi: function() {
        const results = [];
        let _, u2, ref2;
        for (_ = u2 = 0, ref2 = n; 0 <= ref2 ? u2 < ref2 : u2 > ref2; _ = 0 <= ref2 ? ++u2 : --u2) {
          results.push({});
        }
        return results;
      }(),
      g: function() {
        const results = [];
        let _, u2, ref2;
        for (_ = u2 = 0, ref2 = n; 0 <= ref2 ? u2 < ref2 : u2 > ref2; _ = 0 <= ref2 ? ++u2 : --u2) {
          results.push({});
        }
        return results;
      }()
    };
    const update = (m, l) => {
      const k2 = m.j;
      let pi = this.estimate_guesses(m, password);
      if (l > 1) {
        pi *= optimal.pi[m.i - 1][l - 1];
      }
      let g = this.factorial(l) * pi;
      if (!_exclude_additive) {
        g += Math.pow(MIN_GUESSES_BEFORE_GROWING_SEQUENCE, l - 1);
      }
      const ref2 = optimal.g[k2];
      for (const competing_l in ref2) {
        const competing_g = ref2[competing_l];
        if (competing_l > l) {
          continue;
        }
        if (competing_g <= g) {
          return;
        }
      }
      optimal.g[k2][l] = g;
      optimal.m[k2][l] = m;
      return optimal.pi[k2][l] = pi;
    };
    const bruteforce_update = (k2) => {
      const m = make_bruteforce_match(0, k2);
      update(m, 1);
      const results = [];
      let i, u2, ref2;
      for (i = u2 = 1, ref2 = k2; 1 <= ref2 ? u2 <= ref2 : u2 >= ref2; i = 1 <= ref2 ? ++u2 : --u2) {
        const m2 = make_bruteforce_match(i, k2);
        results.push(function() {
          const ref1 = optimal.m[i - 1];
          const results1 = [];
          for (const l in ref1) {
            const last_m = ref1[l];
            const li = parseInt(l);
            if (last_m.pattern === "bruteforce") {
              continue;
            }
            results1.push(update(m2, li + 1));
          }
          return results1;
        }());
      }
      return results;
    };
    const make_bruteforce_match = (i, j) => {
      return {
        pattern: "bruteforce",
        token: password.slice(i, +j + 1 || 9e9),
        i,
        j
      };
    };
    const unwind = (n2) => {
      const optimal_match_sequence2 = [];
      let k2 = n2 - 1;
      let l = void 0;
      let g = Infinity;
      const ref2 = optimal.g[k2];
      for (const candidate_l in ref2) {
        const candidate_g = ref2[candidate_l];
        if (candidate_g < g) {
          l = candidate_l;
          g = candidate_g;
        }
      }
      while (k2 >= 0) {
        const m = optimal.m[k2][l];
        optimal_match_sequence2.unshift(m);
        k2 = m.i - 1;
        l--;
      }
      return optimal_match_sequence2;
    };
    let k, u, ref;
    for (k = u = 0, ref = n; 0 <= ref ? u < ref : u > ref; k = 0 <= ref ? ++u : --u) {
      const ref1 = matches_by_j[k];
      for (let w = 0, len2 = ref1.length; w < len2; w++) {
        const m = ref1[w];
        if (m.i > 0) {
          for (const l in optimal.m[m.i - 1]) {
            const li = parseInt(l);
            update(m, li + 1);
          }
        } else {
          update(m, 1);
        }
      }
      bruteforce_update(k);
    }
    const optimal_match_sequence = unwind(n);
    const optimal_l = optimal_match_sequence.length;
    let guesses;
    if (password.length === 0) {
      guesses = 1;
    } else {
      guesses = optimal.g[n - 1][optimal_l];
    }
    return {
      password,
      guesses,
      guesses_log10: this.log10(guesses),
      sequence: optimal_match_sequence
    };
  },
  estimate_guesses: function(match, password) {
    if (match.guesses != null) {
      return match.guesses;
    }
    let min_guesses = 1;
    if (match.token.length < password.length) {
      min_guesses = match.token.length === 1 ? MIN_SUBMATCH_GUESSES_SINGLE_CHAR : MIN_SUBMATCH_GUESSES_MULTI_CHAR;
    }
    const estimation_functions = {
      bruteforce: this.bruteforce_guesses,
      dictionary: this.dictionary_guesses,
      spatial: this.spatial_guesses,
      repeat: this.repeat_guesses,
      sequence: this.sequence_guesses,
      regex: this.regex_guesses,
      date: this.date_guesses
    };
    const guesses = estimation_functions[match.pattern].call(this, match);
    match.guesses = Math.max(guesses, min_guesses);
    match.guesses_log10 = this.log10(match.guesses);
    return match.guesses;
  },
  bruteforce_guesses: function(match) {
    let guesses = Math.pow(BRUTEFORCE_CARDINALITY, match.token.length);
    if (guesses === Number.POSITIVE_INFINITY) {
      guesses = Number.MAX_VALUE;
    }
    const min_guesses = match.token.length === 1 ? MIN_SUBMATCH_GUESSES_SINGLE_CHAR + 1 : MIN_SUBMATCH_GUESSES_MULTI_CHAR + 1;
    return Math.max(guesses, min_guesses);
  },
  repeat_guesses: function(match) {
    return match.base_guesses * match.repeat_count;
  },
  sequence_guesses: function(match) {
    const first_chr = match.token.charAt(0);
    let base_guesses;
    if (first_chr === "a" || first_chr === "A" || first_chr === "z" || first_chr === "Z" || first_chr === "0" || first_chr === "1" || first_chr === "9") {
      base_guesses = 4;
    } else {
      if (first_chr.match(/\d/)) {
        base_guesses = 10;
      } else {
        base_guesses = 26;
      }
    }
    if (!match.ascending) {
      base_guesses *= 2;
    }
    return base_guesses * match.token.length;
  },
  MIN_YEAR_SPACE: 20,
  REFERENCE_YEAR: new Date().getFullYear(),
  regex_guesses: function(match) {
    const char_class_bases = {
      alpha_lower: 26,
      alpha_upper: 26,
      alpha: 52,
      alphanumeric: 62,
      digits: 10,
      symbols: 33
    };
    let year_space;
    if (match.regex_name in char_class_bases) {
      return Math.pow(char_class_bases[match.regex_name], match.token.length);
    } else {
      switch (match.regex_name) {
        case "recent_year":
          year_space = Math.abs(parseInt(match.regex_match[0]) - this.REFERENCE_YEAR);
          year_space = Math.max(year_space, this.MIN_YEAR_SPACE);
          return year_space;
      }
    }
  },
  date_guesses: function(match) {
    const year_space = Math.max(Math.abs(match.year - this.REFERENCE_YEAR), this.MIN_YEAR_SPACE);
    let guesses = year_space * 365;
    if (match.separator) {
      guesses *= 4;
    }
    return guesses;
  },
  KEYBOARD_AVERAGE_DEGREE: calc_average_degree(adjacency_graphs.qwerty),
  KEYPAD_AVERAGE_DEGREE: calc_average_degree(adjacency_graphs.keypad),
  KEYBOARD_STARTING_POSITIONS: function() {
    const ref = adjacency_graphs.qwerty;
    const results = [];
    for (const k in ref) {
      const v = ref[k];
      results.push(k);
    }
    return results;
  }().length,
  KEYPAD_STARTING_POSITIONS: function() {
    const ref = adjacency_graphs.keypad;
    const results = [];
    for (const k in ref) {
      const v = ref[k];
      results.push(k);
    }
    return results;
  }().length,
  spatial_guesses: function(match) {
    let ref, s, d;
    if ((ref = match.graph) === "qwerty" || ref === "dvorak") {
      s = this.KEYBOARD_STARTING_POSITIONS;
      d = this.KEYBOARD_AVERAGE_DEGREE;
    } else {
      s = this.KEYPAD_STARTING_POSITIONS;
      d = this.KEYPAD_AVERAGE_DEGREE;
    }
    let guesses = 0;
    const L = match.token.length;
    const t = match.turns;
    let i, o, ref1, j, q, ref2, u, ref3;
    for (i = o = 2, ref1 = L; 2 <= ref1 ? o <= ref1 : o >= ref1; i = 2 <= ref1 ? ++o : --o) {
      const possible_turns = Math.min(t, i - 1);
      for (j = q = 1, ref2 = possible_turns; 1 <= ref2 ? q <= ref2 : q >= ref2; j = 1 <= ref2 ? ++q : --q) {
        guesses += this.nCk(i - 1, j - 1) * s * Math.pow(d, j);
      }
    }
    if (match.shifted_count) {
      const S = match.shifted_count;
      const U = match.token.length - match.shifted_count;
      if (S === 0 || U === 0) {
        guesses *= 2;
      } else {
        let shifted_variations = 0;
        for (i = u = 1, ref3 = Math.min(S, U); 1 <= ref3 ? u <= ref3 : u >= ref3; i = 1 <= ref3 ? ++u : --u) {
          shifted_variations += this.nCk(S + U, i);
        }
        guesses *= shifted_variations;
      }
    }
    return guesses;
  },
  dictionary_guesses: function(match) {
    match.base_guesses = match.rank;
    match.uppercase_variations = this.uppercase_variations(match);
    match.l33t_variations = this.l33t_variations(match);
    const reversed_variations = match.reversed && 2 || 1;
    return match.base_guesses * match.uppercase_variations * match.l33t_variations * reversed_variations;
  },
  START_UPPER: /^[A-Z][^A-Z]+$/,
  END_UPPER: /^[^A-Z]+[A-Z]$/,
  ALL_UPPER: /^[^a-z]+$/,
  ALL_LOWER: /^[^A-Z]+$/,
  uppercase_variations: function(match) {
    const word = match.token;
    if (word.match(this.ALL_LOWER) || word.toLowerCase() === word) {
      return 1;
    }
    const ref = [this.START_UPPER, this.END_UPPER, this.ALL_UPPER];
    for (let o = 0, len = ref.length; o < len; o++) {
      const regex = ref[o];
      if (word.match(regex)) {
        return 2;
      }
    }
    const U = function() {
      const ref12 = word.split("");
      const results = [];
      for (let q2 = 0, len1 = ref12.length; q2 < len1; q2++) {
        const chr = ref12[q2];
        if (chr.match(/[A-Z]/)) {
          results.push(chr);
        }
      }
      return results;
    }().length;
    const L = function() {
      const ref12 = word.split("");
      const results = [];
      for (let q2 = 0, len1 = ref12.length; q2 < len1; q2++) {
        const chr = ref12[q2];
        if (chr.match(/[a-z]/)) {
          results.push(chr);
        }
      }
      return results;
    }().length;
    let variations = 0;
    let i, q, ref1;
    for (i = q = 1, ref1 = Math.min(U, L); 1 <= ref1 ? q <= ref1 : q >= ref1; i = 1 <= ref1 ? ++q : --q) {
      variations += this.nCk(U + L, i);
    }
    return variations;
  },
  l33t_variations: function(match) {
    if (!match.l33t) {
      return 1;
    }
    let variations = 1;
    const ref = match.sub;
    for (const subbed in ref) {
      const unsubbed = ref[subbed];
      const chrs = match.token.toLowerCase().split("");
      const S = function() {
        const results = [];
        for (let o = 0, len = chrs.length; o < len; o++) {
          const chr = chrs[o];
          if (chr === subbed) {
            results.push(chr);
          }
        }
        return results;
      }().length;
      const U = function() {
        const results = [];
        for (let o = 0, len = chrs.length; o < len; o++) {
          const chr = chrs[o];
          if (chr === unsubbed) {
            results.push(chr);
          }
        }
        return results;
      }().length;
      if (S === 0 || U === 0) {
        variations *= 2;
      } else {
        const p = Math.min(U, S);
        let possibilities = 0;
        let i, o, ref1;
        for (i = o = 1, ref1 = p; 1 <= ref1 ? o <= ref1 : o >= ref1; i = 1 <= ref1 ? ++o : --o) {
          possibilities += this.nCk(U + S, i);
        }
        variations *= possibilities;
      }
    }
    return variations;
  }
};
export default scoring;
